# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import cimsparql

# -- Project information -----------------------------------------------------

project = "cimsparql"
copyright = "2020, Statnett"
author = "Statnett DataScience <Datascience.Drift@Statnett.no>"

# The full version, including alpha/beta/rc tags
release = cimsparql.__version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.graphviz",
    "sphinx.ext.githubpages",
    "recommonmark",
    "sphinx.ext.imgmath",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "numpy": ("https://docs.scipy.org/doc/numpy", None),
}

autosectionlabel_prefix_document = True
napoleon_include_private_with_doc = True

inheritance_node_attrs = {
    "shape": "folder",
    "fontsize": 14,
    "height": 0.75,
    "fillcolor": "SteelBlue1",
    "style": "filled",
    "arrowType": "open",
    "arrowSize": 1.2,
}

graphviz_output_format = "svg"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

autodoc_default_options = {"special-members": "__call__"}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

html_theme = "sphinx_rtd_theme"

autoclass_content = "both"
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



"""Provide value mappers."""

import pandas as pd


class ValueMapper:
    """Base value mapper."""

    def map(self, df: pd.DataFrame) -> pd.DataFrame:
        return df


class MridMapper(ValueMapper):
    """Mrid mapper.

    Replaces all matches of uri_regex with an empty string in columns which contains mrid in the name
    """

    def __init__(self, uri_regex: str = "^([^#_|^#]+)(#_|#)") -> None:
        self.uri_regex = uri_regex

    def map(self, df: pd.DataFrame) -> pd.DataFrame:  # A003
        for col in filter(lambda col: "mrid" in col, df.columns):
            df[col] = df[col].str.replace(self.uri_regex, "", regex=True)
        return df



import pathlib
from collections.abc import Generator
from contextlib import suppress

import pytest

import tests.t_utils.common as t_common
import tests.t_utils.entsoe_models as t_entsoe

this_dir = pathlib.Path(__file__).parent


@pytest.fixture(scope="session", autouse=True)
def delete_models() -> Generator[None, None, None]:
    """Fixture for deleting micro model repos created in during tests."""
    yield
    models = t_entsoe.micro_models() + t_entsoe.smallgrid_models()
    for test_model in filter(lambda tm: tm.model is not None and tm.cleanup, models):
        assert test_model.model
        for client in test_model.model.distinct_clients:
            with suppress(Exception):
                client.delete_repo()


@pytest.fixture(scope="session", autouse=True)
def delete_picasso_repo() -> Generator[None, None, None]:
    yield
    with suppress(Exception):
        client = t_common.initialized_rdf4j_repo()
        client.delete_repo()



import functools
import logging
import os

import tests.t_utils.common as t_common
from cimsparql.graphdb import GraphDBClient, ServiceConfig
from cimsparql.model import ModelConfig, get_federated_cim_model

logger = logging.getLogger()


def combined_graphdb_service() -> ServiceConfig:
    repo = os.getenv("GRAPHDB_COMBINED_REPO", "abot_combined")
    return ServiceConfig(repo=repo, max_delay_seconds=1, token=os.getenv("GRAPHDB_TOKEN"))


@functools.lru_cache
def federated_model() -> t_common.ModelTest:
    if os.getenv("CI") or os.getenv("GRAPHDB_TOKEN") is None:
        return t_common.ModelTest(None, must_run_in_ci=False, cleanup=False)
    eq_repo = os.getenv("GRAPHDB_EQ", "abot-eq-1")
    system_state_repo = os.getenv("GRAPHDB_STATE", "abot-situations-1")
    server = os.getenv("GRAPHDB_SERVER", "modelstore.dev.form.statnett.no:8443")
    eq_client_cfg = ServiceConfig(eq_repo, server=server, max_delay_seconds=1)
    tpsvssh_client_cfg = ServiceConfig(system_state_repo, server=server, max_delay_seconds=1)
    eq_client = GraphDBClient(eq_client_cfg)
    tpsvssh_client = GraphDBClient(tpsvssh_client_cfg)

    m_cfg = ModelConfig(f"repository:{system_state_repo},infer=false", f"repository:{eq_repo},infer=false")
    model = None
    try:
        model = get_federated_cim_model(eq_client, tpsvssh_client, m_cfg)
    except Exception:
        logger.exception("Failed to get federated model")
    return t_common.ModelTest(model, must_run_in_ci=False, cleanup=False)


def all_custom_models() -> list[t_common.ModelTest]:
    return [federated_model()]



"""Implementation of an in-memory GraphDBClient using pyoxigraph as engine."""

import re
from pathlib import Path

import httpx
from pyoxigraph import NamedNode, QueryResultsFormat, QuerySolutions, RdfFormat, Store

from cimsparql.graphdb import GraphDBClient, default_namespaces
from cimsparql.sparql_result_json import SparqlResultJson


class InvalidResultTypeError(RuntimeError):
    def __init__(self, result_type: str) -> None:
        super().__init__(f"Only queries resulting a solution of type 'QuerySolutions' are supported. Got {result_type}")


class LocalClient(GraphDBClient):
    def __init__(self, store: Store | None = None, strip_service_specifier: bool = False) -> None:
        self.strip_service_specifier = strip_service_specifier
        self.store = store or Store()
        super().__init__()

    def exec_query(self, query: str) -> SparqlResultJson:
        if self.strip_service_specifier:
            query = re.sub("SERVICE[^{]+", "", query)

        result = self.store.query(query, prefixes=self.prefixes)
        if not isinstance(result, QuerySolutions):
            raise InvalidResultTypeError(str(type(result)))

        serialized_result = result.serialize(format=QueryResultsFormat.JSON)
        assert serialized_result
        return SparqlResultJson.model_validate_json(serialized_result)

    def get_prefixes(self, http_transport: httpx.BaseTransport | None = None) -> dict[str, str]:
        _ = http_transport
        if self._prefixes is None:
            self._prefixes = default_namespaces()
        return self._prefixes

    def delete_repo(self) -> None:
        raise NotImplementedError(
            "LocalClient supports only one repository. Create a new client instead of deleting the repo"
        )

    def upload_rdf(self, content: Path | bytes, rdf_format: str, params: dict[str, str] | None = None) -> None:
        data_format = oxigraph_format(rdf_format)
        params = params or {}
        named_graph = NamedNode(params["graph"]) if "graph" in params else None
        base_iri = params.get("base_iri")
        if isinstance(content, Path):
            self.store.load(path=content, format=data_format, to_graph=named_graph, base_iri=base_iri)
        else:
            self.store.load(content, format=data_format, to_graph=named_graph, base_iri=base_iri)

    def update_query(self, query: str) -> None:
        _ = query
        raise NotImplementedError("Update query is currently not implemented")

    def get_namespace(self, prefix: str) -> str:
        return self.prefixes[prefix]

    def set_namespace(self, prefix: str, value: str) -> None:
        if self._prefixes is None:
            self._prefixes = {}
        self._prefixes[prefix] = value


def oxigraph_format(fmt: str) -> RdfFormat:
    format_map = {
        "rdf/xml": RdfFormat.RDF_XML,
        "n-triples": RdfFormat.N_TRIPLES,
        "turtle": RdfFormat.TURTLE,
        "n3": RdfFormat.N3,
        "n-quads": RdfFormat.N_QUADS,
        "json-ld": RdfFormat.JSON_LD,
        "trig": RdfFormat.TRIG,
    }
    if fmt not in format_map:
        raise ValueError(f"format must be one of {', '.join(format_map.keys())}")
    return format_map[fmt]



from logging import INFO
from typing import Any

import pytest
import tenacity
from SPARQLWrapper import SPARQLWrapper

from cimsparql.graphdb import GraphDBClient, ServiceConfig
from cimsparql.retry_cb import RetryCallback
from cimsparql.sparql_result_json import SparqlResultJsonFactory


def test_query_name_registered_in_pre_call():
    retry_callback = RetryCallback()
    retry_callback.pre_call("# Name: some random name\nselect * where {?s ?p ?o}")
    assert retry_callback.query_name == "some random name"


class FailFirstSparqlWrapper(SPARQLWrapper):
    num_calls = 0

    def queryAndConvert(self) -> dict[str, Any]:  # noqa: N802
        self.num_calls += 1
        if self.num_calls == 1:
            raise ValueError("Something always goes wrong on the first attempt")
        return SparqlResultJsonFactory.build().model_dump(mode="json")


def test_after_callback(caplog: pytest.LogCaptureFixture):
    wrapper = FailFirstSparqlWrapper("http://some-sparql-endpoint")
    client = GraphDBClient(
        service_cfg=ServiceConfig(retry_stop_criteria=tenacity.stop_after_attempt(2)),
        sparql_wrapper=wrapper,
    )
    with caplog.at_level(INFO):
        client.exec_query("# Name: nSelect everything\nselect * where {?s ?p ?o}")

    # Expect one message to contain be logged with the query name
    substrings = ["Select everything:", "Something always goes wrong"]
    assert sum(all(s in msg for s in substrings) for msg in caplog.messages) == 1



from __future__ import annotations

import dataclasses
import os
from pathlib import Path
from string import Template
from typing import TYPE_CHECKING

import pytest

import tests.t_utils.common as t_common
from cimsparql.graphdb import GraphDBClient, RestApi

if TYPE_CHECKING:
    from collections.abc import Generator


def skip_rdf4j_test(rdf4j_gdb: GraphDBClient | None) -> bool:
    # On CI in GitHub these tests should always run
    return not (rdf4j_gdb or os.getenv("CI"))


@pytest.fixture(scope="module")
def rdf4j_gdb() -> GraphDBClient:
    try:
        return t_common.initialized_rdf4j_repo()
    except Exception as exc:
        if os.getenv("CI"):
            pytest.fail(f"{exc}")
        else:
            pytest.skip(f"{exc}")


@pytest.mark.parametrize(
    ("query", "expect"),
    [
        ("select ?name where {ex:Picasso foaf:firstName ?name}", [{"name": "Pablo"}]),
        (
            "select ?city {ex:Picasso ex:homeAddress ?address . ?address ex:city ?city}",
            [{"city": "Madrid"}],
        ),
    ],
)
def test_rdf4j_picasso_data(rdf4j_gdb: GraphDBClient, query: str, expect: list[dict[str, str]]):
    prefixes = Template("PREFIX ex:<${ex}>\nPREFIX foaf:<${foaf}>").substitute(rdf4j_gdb.prefixes)
    result = rdf4j_gdb.exec_query(f"{prefixes}\n{query}").results.values_as_dict()
    assert result == expect


def test_rdf4j_prefixes(rdf4j_gdb: GraphDBClient):
    assert set(rdf4j_gdb.prefixes.keys()).issuperset({"ex", "foaf"})


@pytest.fixture
def upload_client() -> Generator[GraphDBClient, None, None]:
    client = None
    try:
        yield t_common.init_repo_rdf4j(t_common.rdf4j_url(), "upload")
    except Exception as exc:
        if os.getenv("CI"):
            pytest.fail(f"{exc}")
        else:
            pytest.skip(f"{exc}")
    finally:
        if client:
            client.delete_repo()


def test_upload_rdf_xml(upload_client: GraphDBClient):
    xml_file = Path(__file__).parent / "data/demo.xml"
    upload_client.upload_rdf(xml_file, "rdf/xml")

    prefixes = Template("PREFIX rdf:<${rdf}>\nPREFIX md:<${md}>").substitute(upload_client.prefixes)
    sparql_result = upload_client.exec_query(f"{prefixes}\nSELECT * WHERE {{?s rdf:type md:FullModel}}")
    assert len(sparql_result.results.bindings) == 1


def test_get_table_default_arg(rdf4j_gdb: GraphDBClient):
    data = rdf4j_gdb.get_table("SELECT * {?s ?o ?p}")[0]
    assert len(data) == 6


def test_namespaces(rdf4j_gdb: GraphDBClient):
    ns = "http://mynamepace.org"
    rdf4j_gdb.set_namespace("myns", ns)
    fetched_ns = rdf4j_gdb.get_namespace("myns")
    assert ns == fetched_ns


def test_upload_with_context(upload_client: GraphDBClient):
    xml_file = Path(__file__).parent / "data/demo.xml"
    graph = "<http://mygraph.com/demo/1/1>"
    upload_client.upload_rdf(xml_file, "rdf/xml", {"context": graph})

    prefixes = Template("PREFIX rdf:<${rdf}>\nPREFIX md:<${md}>").substitute(upload_client.prefixes)
    sparql_result = upload_client.exec_query(
        f"{prefixes}\nSELECT * WHERE {{GRAPH {graph} {{?s rdf:type md:FullModel}}}}"
    )
    assert len(sparql_result.results.bindings) == 1


def test_direct_sparql_endpoint(rdf4j_gdb: GraphDBClient):
    service_cfg_direct = dataclasses.replace(
        rdf4j_gdb.service_cfg,
        server=rdf4j_gdb.service_cfg.url,
        rest_api=RestApi.DIRECT_SPARQL_ENDPOINT,
    )

    gdb_direct = GraphDBClient(service_cfg_direct)
    query = "SELECT * {?s ?p ?o}"
    res_direct = gdb_direct.exec_query(query)
    res_rdf4j = rdf4j_gdb.exec_query(query)
    assert res_direct == res_rdf4j



from pathlib import Path

import pytest

from cimsparql.adaptions import XmlModelAdaptor, is_uuid


@pytest.fixture
def xml_adaptor():
    folder = Path(__file__).parent / "data" / "micro"
    return XmlModelAdaptor.from_folder(folder)


def test_namespaces(xml_adaptor: XmlModelAdaptor):
    ns = xml_adaptor.namespaces()
    assert {"xsd", "cim"}.issubset(set(ns.keys()))


def test_add_mrid(xml_adaptor: XmlModelAdaptor):
    query = "select * where {graph ?g {?s cim:IdentifiedObject.mRID ?mrid}}"
    assert len(list(xml_adaptor.select_query(query, prefixes=xml_adaptor.namespaces()))) == 0
    xml_adaptor.add_mrid()
    assert len(list(xml_adaptor.select_query(query, prefixes=xml_adaptor.namespaces()))) > 0


def test_tpsvssh_contexts(xml_adaptor: XmlModelAdaptor):
    assert len(set(xml_adaptor.tpsvssh_contexts())) == 6


@pytest.mark.parametrize(
    ("value", "result"),
    [
        ("17086487-56ba-4979-b8de-064025a6b4da", True),
        ("1708648756ba-4979-b8de-064025a6b4da", False),
    ],
)
def test_is_uuid(value: str, result: bool):
    assert is_uuid(value) is result


def test_add_internal_eq_link(xml_adaptor: XmlModelAdaptor):
    xml_adaptor.add_internal_eq_link("http://eq.com")
    query = f"select * where {{graph ?g {{?node <{XmlModelAdaptor.eq_predicate}> ?eq_uri}}}}"
    assert len(list(xml_adaptor.select_query(query))) == 1


def test_add_sv_injection(xml_adaptor: XmlModelAdaptor) -> None:
    query = "select * where {graph ?g {?s a cim:SvInjection}}"
    orig_number = len(list(xml_adaptor.select_query(query, prefixes=xml_adaptor.namespaces())))
    xml_adaptor.add_zero_sv_injection()
    assert len(list(xml_adaptor.select_query(query, prefixes=xml_adaptor.namespaces()))) > orig_number


def test_add_eic_code(xml_adaptor: XmlModelAdaptor) -> None:
    query = "select * where {graph ?g {?s a cim:Substation}}"
    num_substations = len(list(xml_adaptor.select_query(query, prefixes=xml_adaptor.namespaces())))
    assert num_substations > 0
    xml_adaptor.add_eic_code()
    query = "select * where {graph ?g {?s entsoeSecretariat:IdentifiedObject.energyIdentCodeEIC ?eic}}"
    assert len(list(xml_adaptor.select_query(query, prefixes=xml_adaptor.namespaces()))) == num_substations


def test_add_network_analysis_enable(xml_adaptor: XmlModelAdaptor) -> None:
    xml_adaptor.add_network_analysis_enable()
    query = """
        select * where {graph ?g {?terminal cim:Terminal.ConductingEquipment/SN:Equipment.networkAnalysisEnable true }}
        """
    assert len(list(xml_adaptor.select_query(query, prefixes=xml_adaptor.namespaces()))) > 0


def test_add_generating_unit(xml_adaptor: XmlModelAdaptor) -> None:
    xml_adaptor.add_generating_unit()
    query = "select * {graph ?g {?machine cim:SynchronousMachine.GeneratingUnit ?unit}}"
    assert len(list(xml_adaptor.select_query(query, prefixes=xml_adaptor.namespaces()))) > 0


def test_add_market_code_to_non_conform_load(xml_adaptor: XmlModelAdaptor) -> None:
    query = "select * {graph ?g {?group SN:NonConformLoadGroup.ScheduleResource/SN:ScheduleResource.marketCode ?code}}"
    assert len(list(xml_adaptor.select_query(query, prefixes=xml_adaptor.namespaces()))) == 0

    xml_adaptor.add_market_code_to_non_conform_load()
    assert len(list(xml_adaptor.select_query(query, prefixes=xml_adaptor.namespaces()))) == 6


def test_num_base_voltage_in_serialized_graph(xml_adaptor: XmlModelAdaptor) -> None:
    result = xml_adaptor.nq_bytes(set(xml_adaptor.eq_contexts())).decode()
    assert result.count("BaseVoltage.nominalVoltage") == 8


def test_add_ras(xml_adaptor: XmlModelAdaptor) -> None:
    xml_adaptor.add_protective_action_equipment()
    query = "select * {graph ?g {?s a ALG:ProtectiveActionEquipment}}"
    assert len(list(xml_adaptor.select_query(query, prefixes=xml_adaptor.namespaces()))) == 1



from pathlib import Path

import pytest

from cimsparql.local_client import InvalidResultTypeError, LocalClient, oxigraph_format


def nl_eq_path() -> Path:
    return Path(__file__).parent / "data/micro/20171002T0930Z_NL_EQ_3.xml"


@pytest.mark.parametrize("pass_bytes", [True, False])
def test_upload_rdf(pass_bytes: bool):
    data_path = nl_eq_path()
    params = {"base_iri": "http://example.com/"}
    client = LocalClient()
    if pass_bytes:
        with data_path.open("rb") as infile:
            content = infile.read()
        client.upload_rdf(content, "rdf/xml", params)
    else:
        client.upload_rdf(data_path, "rdf/xml", params)

    result = client.exec_query("SELECT (count(*) as ?num) WHERE {?s ?p ?o}")
    assert int(result.results.bindings[0]["num"].value) > 0


def test_upload_to_named_graph():
    data_path = nl_eq_path()
    client = LocalClient()
    params = {"base_iri": "http://base.com/", "graph": "http://nl-eq-graph"}
    client.upload_rdf(data_path, "rdf/xml", params)
    assert {str(node) for node in client.store.named_graphs()} == {"<http://nl-eq-graph>"}


def test_error_on_wrong_query_type():
    client = LocalClient()
    with pytest.raises(InvalidResultTypeError):
        client.exec_query("ASK {?s ?p ?o}")


def test_delete_repo_not_implemented():
    with pytest.raises(NotImplementedError):
        LocalClient().delete_repo()


def test_get_prefixes():
    assert len(LocalClient().get_prefixes()) > 0


def test_update_query():
    with pytest.raises(NotImplementedError):
        LocalClient().update_query("")


def test_namespaces_round_trip():
    client = LocalClient()
    client.set_namespace("ns", "http://namespace1")
    assert client.get_namespace("ns") == "http://namespace1"


def test_oxigraph_raises_on_unknown():
    with pytest.raises(ValueError, match="format must be one of"):
        oxigraph_format("json-ldd")



"""The cimsparql.model module contains the base class Model."""

from __future__ import annotations

import functools
import logging
import re
import time
from collections import defaultdict
from dataclasses import dataclass, field
from functools import cached_property
from typing import TYPE_CHECKING, ParamSpec, Self, TypeVar
from uuid import uuid4

import pandas as pd

from cimsparql import templates
from cimsparql.data_models import (
    AcLinesDataFrame,
    AssociatedSwitchesDataFrame,
    BaseVoltageDataFrame,
    BordersDataFrame,
    BranchComponentDataFrame,
    BranchWithdrawDataFrame,
    BusDataFrame,
    ConnectionsDataFrame,
    ConnectivityNodeDataFrame,
    ConvertersDataFrame,
    CoordinatesDataFrame,
    DcActiveFlowDataFrame,
    DisconnectedDataFrame,
    ExchangeDataFrame,
    FullModelDataFrame,
    GenUnitAndSyncMachineMridDataFrame,
    HVDCBidzonesDataFrame,
    HVDCDataFrame,
    LoadsDataFrame,
    MarketDatesDataFrame,
    PhaseTapChangerDataFrame,
    PowerFlowDataFrame,
    RASEquipmentDataFrame,
    RegionsDataFrame,
    StationGroupCodeNameDataFrame,
    StationGroupForPowerUnitDataFrame,
    SubstationVoltageDataFrame,
    SvInjectionDataFrame,
    SvPowerDeviationDataFrame,
    SwitchesDataFrame,
    SynchronousMachinesDataFrame,
    TransfConToConverterDataFrame,
    TransformersDataFrame,
    TransformerWindingDataFrame,
    TransformerWindingsDataFrame,
    WindGeneratingUnitsDataFrame,
)
from cimsparql.graphdb import GraphDBClient, ServiceConfig
from cimsparql.type_mapper import TypeMapper
from cimsparql.utils import query_name

if TYPE_CHECKING:
    from collections.abc import Callable, Generator, Iterable
    from string import Template
    from types import TracebackType

    from cimsparql.sparql_result_json import SparqlResultValue
    from cimsparql.value_mapper import ValueMapper


@dataclass
class ModelConfig:
    system_state_repo: str | None = None
    eq_repo: str | None = None
    value_mappers: Iterable[ValueMapper] = field(default_factory=list)


logger = logging.getLogger()
T = TypeVar("T")
P = ParamSpec("P")


def time_it(f: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(f)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        started = time.time()
        result = f(*args, **kwargs)
        finished = time.time()
        logger.debug("%s took %f seconds", f.__name__, finished - started)
        return result

    return wrapped


class Model:
    def __init__(
        self,
        clients: dict[str, GraphDBClient],
        config: ModelConfig | None = None,
        mapper: TypeMapper | None = None,
    ) -> None:
        self.clients = clients
        self.config = config or ModelConfig()
        self.mapper = mapper or TypeMapper(self.get_client("Type mapper"))

    @property
    def distinct_clients(self) -> list[GraphDBClient]:
        obj_ids = set[int]()
        distinct = list[GraphDBClient]()
        for client in self.clients.values():
            if id(client) not in obj_ids:
                obj_ids.add(id(client))
                distinct.append(client)
        return distinct

    def get_client(self, query_name: str) -> GraphDBClient:
        """Return the correct graph db client to execute a query.

        By default there is only one client so the same client is returned in all cases
        """
        return self.clients[query_name]

    def __enter__(self) -> Self:
        transaction_id = str(uuid4())
        for client in self.clients.values():
            client.add_correlation_id_to_header(transaction_id)
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        _, _, _ = exc_type, exc, exc_tb
        for client in self.clients.values():
            client.clear_correlation_id_from_header()

    @property
    def client(self) -> GraphDBClient:
        return self.get_client("default")

    @client.setter
    def client(self, default_client: GraphDBClient) -> None:
        self.clients["default"] = default_client

    @staticmethod
    def _col_map(data_row: dict[str, SparqlResultValue]) -> dict[str, str]:
        return {column: data.datatype if data.datatype else data.value_type for column, data in data_row.items()}

    @classmethod
    def col_map(cls, data_row: dict[str, SparqlResultValue], columns: dict[str, str]) -> dict[str, str]:
        columns = columns or {}
        col_map = cls._col_map(data_row)
        col_map.update(columns)
        return col_map

    @property
    def map_data_types(self) -> bool:
        pref = self.client.prefixes
        return "cim" in pref and self.mapper.have_cim_version(pref["cim"])

    def _convert_result(
        self,
        result: pd.DataFrame,
        data_row: dict[str, SparqlResultValue],
        index: str | None = None,
        columns: dict[str, str] | None = None,
    ) -> pd.DataFrame:
        col_map = self.col_map(data_row, columns or {})
        result = self.mapper.map_data_types(result, col_map)
        for v_mapper in self.config.value_mappers:
            result = v_mapper.map(result)

        if index:
            return result.set_index(index)
        return result

    def get_table_and_convert(
        self, query: str, index: str | None = None, columns: dict[str, str] | None = None
    ) -> pd.DataFrame:
        name = query_name(query)
        client = self.get_client(name)
        result, data_row = client.get_table(query)
        return self._convert_result(result, data_row, index, columns)

    def template_to_query(self, template: Template, substitutes: dict[str, str] | None = None) -> str:
        """Convert provided template to query."""
        substitutes = substitutes or {}

        # Extract name from the query. We don't need to perform actual substitutions to do this,
        # we just let placeholder be left
        name = query_name(template.safe_substitute())
        client = self.get_client(name)
        state_repo = self.config.system_state_repo or client.service_cfg.url
        eq_repo = self.config.eq_repo or client.service_cfg.url

        return template.safe_substitute({"repo": state_repo, "eq_repo": eq_repo} | substitutes | self.client.prefixes)

    @cached_property
    def cim_version(self) -> int:
        if m := re.search("cim(\\d+)", self.client.prefixes["cim"]):
            return int(m.group(1))
        return 0

    @property
    def full_model_query(self) -> str:
        return self.template_to_query(templates.FULL_MODEL_QUERY)

    @time_it
    def full_model(self) -> FullModelDataFrame:
        """Return all models where all dependencies has been created and is available.

        All profiles EQ/SSH/TP/SV will define a md:FullModel with possible dependencies. One profile
        could be dependent on more than one other. This function will return the models for SSH/TP
        and SV that is available from current repo or provided <repo>.

        Example:
        >> model.full_model()
        """
        df = self.get_table_and_convert(self.full_model_query)
        return FullModelDataFrame(df)

    @property
    def market_dates_query(self) -> str:
        """Market activation date for this repository."""
        return self.template_to_query(templates.MARKET_DATES_QUERY)

    @time_it
    def market_dates(self) -> MarketDatesDataFrame:
        df = self.get_table_and_convert(self.market_dates_query, index="mrid")
        return MarketDatesDataFrame(df)

    def bus_data_query(self, region: str | None = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.BUS_DATA_QUERY, substitutes)

    def transformer_center_nodes_query(self, region: str | None = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.TRANSFORMER_CENTER_NODES_QUERY, substitutes)

    @time_it
    def bus_data(self, region: str | None = None) -> BusDataFrame:
        """Query name of topological nodes (TP query).

        Args:
           region: Limit to region (use None to get all)
        """
        dfs = [
            self.get_table_and_convert(self.bus_data_query(region), index="node"),
            self.transformer_center_nodes(region),
        ]
        df = pd.concat(dfs)
        return BusDataFrame(df)

    def loads_query(self, region: str | None = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.LOADS_QUERY, substitutes)

    @time_it
    def transformer_center_nodes(self, region: str | None = None) -> BusDataFrame:
        df = self.get_table_and_convert(self.transformer_center_nodes_query(region), index="node")
        return BusDataFrame(df)

    @time_it
    def loads(self, region: str | None = None) -> LoadsDataFrame:
        """Query load data.

        Args:
           region: regexp that limits to region
        """
        query = self.loads_query(region)
        df = self.get_table_and_convert(query, index="mrid")
        return LoadsDataFrame(df)

    def wind_generating_units_query(self, region: str | None = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.WIND_GENERATING_UNITS_QUERY, substitutes)

    @time_it
    def wind_generating_units(self, region: str | None = None) -> WindGeneratingUnitsDataFrame:
        """Query wind generating units.

        Args:
           region: regexp that limits to region

        Example:
            >>> from cimsparql.model import get_single_client_model
            >>> server_url = "127.0.0.1:7200"
            >>> model = get_single_client_model(server_url, "LATEST")
            >>> model.wind_generating_units()
        """
        query = self.wind_generating_units_query(region)
        df = self.get_table_and_convert(query, index="mrid")
        return WindGeneratingUnitsDataFrame(df)

    def synchronous_machines_query(self, region: str | None = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.SYNCHRONOUS_MACHINES_QUERY, substitutes)

    @time_it
    def synchronous_machines(self, region: str | None = None) -> SynchronousMachinesDataFrame:
        query = self.synchronous_machines_query(region)
        df = self.get_table_and_convert(query, index="mrid")
        return SynchronousMachinesDataFrame(df)

    def connections_query(self, region: str | None = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.CONNECTIONS_QUERY, substitutes)

    @time_it
    def connections(self, region: str | None = None) -> ConnectionsDataFrame:
        """Query connectors.

        Args:
           region: regexp that limits to region

        Example:
            >>> from cimsparql.model import get_single_client_model
            >>> server_url = "127.0.0.1:7200"
            >>> model = get_single_client_model(server_url, "LATEST")
            >>> model.connections()
        """
        query = self.connections_query(region)
        df = self.get_table_and_convert(query, index="mrid")
        return ConnectionsDataFrame(df)

    def borders_query(self, region: str | None = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.BORDERS_QUERY, substitutes)

    @time_it
    def borders(self, region: str | None = None) -> BordersDataFrame:
        """Retrieve ACLineSegments where one terminal is inside and the other is outside the region.

        Args:
            region: Inside area
            limit: return first 'limit' number of rows
        """
        query = self.borders_query(region)
        df = self.get_table_and_convert(query, index="mrid")
        return BordersDataFrame(df)

    def exchange_query(self, region: str | None = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.EXCHANGE_QUERY, substitutes)

    @time_it
    def exchange(self, region: str | None = None) -> ExchangeDataFrame:
        """Retrieve ACLineSegments where one terminal is inside and the other is outside the region.

        Args:
            region: Inside area
        """
        if region is None:
            cols = ["name", "node", "status", "p", "market_code"]
            index = pd.Index([], name="mrid")
            df = pd.DataFrame([], columns=cols, index=index)
            return ExchangeDataFrame(df)
        query = self.exchange_query(region)
        df = self.get_table_and_convert(query, index="mrid")
        return ExchangeDataFrame(df)

    def converters_query(self, region: str | None = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.CONVERTERS_QUERY, substitutes)

    @time_it
    def converters(self, region: str | None = None) -> ConvertersDataFrame:
        query = self.converters_query(region)
        df = self.get_table_and_convert(query, index="mrid")
        return ConvertersDataFrame(df)

    def transformers_connected_to_converter_query(self, region: str | None = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.TRANSFORMERS_CONNECTED_TO_CONVERTER_QUERY, substitutes)

    @time_it
    def transformers_connected_to_converter(self, region: str | None = None) -> TransfConToConverterDataFrame:
        """Query list of transformer connected at a converter (Voltage source or DC).

        Args:
           region: regexp that limits to region

        """
        query = self.transformers_connected_to_converter_query(region)
        df = self.get_table_and_convert(query, index="mrid")
        return TransfConToConverterDataFrame(df)

    def ac_lines_query(self, region: str | None = None, rate: str | None = None) -> str:
        substitutes = {"region": region or ".*", "rate": rate or "Normal@20"}
        return self.template_to_query(templates.AC_LINE_QUERY, substitutes)

    @time_it
    def ac_lines(self, region: str | None = None, rate: str | None = None) -> AcLinesDataFrame:
        """Query ac line segments.

        Args:
           region: regexp that limits to region
           rate: specific rate (default: Normal@20)

        Example:
            >>> from cimsparql.model import get_single_client_model
            >>> server_url = "127.0.0.1:7200"
            >>> model = get_single_client_model(server_url, "LATEST")
            >>> model.ac_lines()
        """
        query = self.ac_lines_query(region, rate)
        df = self.get_table_and_convert(query, index="mrid")
        return AcLinesDataFrame(df)

    def series_compensators_query(self, region: str | None = None, rate: str | None = None) -> str:
        substitutes = {"region": region or ".*", "rate": rate or "Normal@20"}
        return self.template_to_query(templates.SERIES_COMPENSATORS_QUERY, substitutes)

    @time_it
    def series_compensators(self, region: str | None = None, rate: str | None = None) -> BranchComponentDataFrame:
        """Query series compensators.

        Args:
           region: regexp that limits to region
           rate: specific rate (default: Normal@20)
        """
        query = self.series_compensators_query(region, rate)
        df = self.get_table_and_convert(query, index="mrid")
        return BranchComponentDataFrame(df)

    def transformers_query(self, region: str | None = None, rate: str | None = None) -> str:
        substitutes = {"region": region or ".*", "rate": rate or "Normal@20"}
        return self.template_to_query(templates.TRANSFORMERS_QUERY, substitutes)

    @time_it
    def transformers(self, region: str | None = None, rate: str | None = None) -> TransformersDataFrame:
        """Query transformer windings.

        Args:
           region: regexp that limits to region
           rate: specific rate (default: Normal@20)
        """
        query = self.transformers_query(region, rate)
        df = self.get_table_and_convert(query)
        return TransformersDataFrame(df)

    @property
    def winding_angle_query(self) -> str:
        return self.template_to_query(templates.TRANSFORMER_WINDING_ANGLE_QUERY)

    def winding_angle(self) -> pd.DataFrame:
        return self.get_table_and_convert(self.winding_angle_query, index="mrid")

    @property
    def winding_query(self) -> str:
        return self.template_to_query(templates.WINDING)

    def winding_loss_query(self, region: str | None = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.WINDING_LOSS_QUERY, substitutes)

    def transformer_branches_query(self, region: str | None = None, rate: str | None = None) -> str:
        substitutes = {"region": region or ".*", "rate": rate or "Normal@20"}
        return self.template_to_query(templates.TRANSFORMER_BRANCHES_QUERY, substitutes)

    @time_it
    def transformer_branches(self, region: str | None = None, rate: str | None = None) -> TransformerWindingDataFrame:
        """Query transformer branches.

        For two winding transformers will give two "branches" (x are nodes, o is the fictisous center node)

        x --- o ---- x

        Three winding transformers

            x
            |
        x - o - x



        Example:
            >>> from cimsparql.model import get_single_client_model
            >>> server_url = "127.0.0.1:7200"
            >>> model = get_single_client_model(server_url, "LATEST")
            >>> model.transformer_branches()
        """
        query = self.transformer_branches_query(region, rate)
        data = self.get_table_and_convert(query, index="mrid")
        query_loss = self.winding_loss_query(region)
        loss = self.get_table_and_convert(query_loss, index="mrid")
        df = data.assign(ploss_1=0.0, ploss_2=lambda df: df["connectivity_node_2"].map(loss["ploss_2"]))
        data.update(self.winding_angle())
        return TransformerWindingDataFrame(df)

    @property
    def substation_voltage_level_query(self) -> str:
        return self.template_to_query(templates.SUBSTATION_VOLTAGE_LEVEL_QUERY)

    def substation_voltage_level(self) -> SubstationVoltageDataFrame:
        query = self.substation_voltage_level_query
        df = self.get_table_and_convert(query, index="substation")
        return SubstationVoltageDataFrame(df)

    @property
    def disconnected_query(self) -> str:
        return self.template_to_query(templates.DISCONNECTED_QUERY)

    @time_it
    def disconnected(self) -> DisconnectedDataFrame:
        """Query disconnected status from ssh profile (not available in GraphDB)."""
        df = self.get_table_and_convert(self.disconnected_query)
        return DisconnectedDataFrame(df)

    @property
    def powerflow_query(self) -> str:
        return self.template_to_query(templates.POWER_FLOW_QUERY)

    @time_it
    def powerflow(self) -> PowerFlowDataFrame:
        """Query powerflow from sv profile (not available in GraphDB)."""
        df = self.get_table_and_convert(self.powerflow_query, index="mrid")
        return PowerFlowDataFrame(df)

    @property
    def phase_tap_changer_query(self) -> str:
        return self.template_to_query(templates.PHASE_TAP_CHANGER)

    def phase_tap_changer(self) -> PhaseTapChangerDataFrame:
        return PhaseTapChangerDataFrame(self.get_table_and_convert(self.phase_tap_changer_query))

    @property
    def transformer_windings_query(self) -> str:
        return self.template_to_query(templates.TRANSFORMER_WINDINGS_QUERY)

    @time_it
    def transformer_windings(self) -> TransformerWindingsDataFrame:
        """Query windings from EQ profile."""
        df = self.get_table_and_convert(self.transformer_windings_query, index="w_mrid")
        return TransformerWindingsDataFrame(df)

    @property
    def coordinates_query(self) -> str:
        return self.template_to_query(templates.COORDINATES_QUERY)

    @time_it
    def coordinates(self) -> CoordinatesDataFrame:
        return CoordinatesDataFrame(self.get_table_and_convert(self.coordinates_query))

    @property
    def st_group_codes_names_query(self) -> str:
        return self.template_to_query(templates.STATION_GROUP_CODE_NAME_QUERY)

    @time_it
    def station_group_codes_and_names(self) -> StationGroupCodeNameDataFrame:
        df = self.get_table_and_convert(self.st_group_codes_names_query).set_index("station_group")
        if df.index.has_duplicates:
            duplicate_indices = df.index[df.index.duplicated(keep="first")]
            logger.warning("Found duplicated names for following station groups: %s", df.loc[duplicate_indices])

        return StationGroupCodeNameDataFrame(df[~df.index.duplicated(keep="first")])

    def branch_node_withdraw_query(self, region: str | None = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.BRANCH_NODE_WITHDRAW_QUERY, substitutes)

    @time_it
    def branch_node_withdraw(self, region: str | None = None) -> BranchWithdrawDataFrame:
        """Query branch flow from sv profile."""
        query = self.branch_node_withdraw_query(region)
        df = self.get_table_and_convert(query, index="mrid")
        return BranchWithdrawDataFrame(df)

    def dc_active_flow_query(self, region: str | None = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.DC_ACTIVE_POWER_FLOW_QUERY, substitutes)

    @time_it
    def dc_active_flow(self, region: str | None = None) -> DcActiveFlowDataFrame:
        query = self.dc_active_flow_query(region)
        df = self.get_table_and_convert(query).astype({"p": float})
        # Unable to group on max within the sparql query so we do it here.
        df = df.iloc[df.groupby("mrid")["p"].idxmax()].set_index("mrid")
        df["p"] *= df["direction"]
        return DcActiveFlowDataFrame(df.drop(columns="direction"))

    @property
    def sv_injection_query(self) -> str:
        return self.template_to_query(templates.SV_INJECTION_QUERY)

    @time_it
    def sv_injection(self) -> SvInjectionDataFrame:
        df = self.get_table_and_convert(self.sv_injection_query)
        return SvInjectionDataFrame(df)

    @property
    def regions_query(self) -> str:
        return self.template_to_query(templates.REGIONS_QUERY)

    @time_it
    def regions(self) -> RegionsDataFrame:
        """Query regions.

        Property:
           regions: List of regions in database

        Example:
            >>> from cimsparql.model import get_single_client_model
            >>> server_url = "127.0.0.1:7200"
            >>> model = get_single_client_model(server_url, "LATEST")
            >>> model.regions
        """
        # TODO: Probably deprecate this property in the future. But keep for now.
        # We need to find a solution for custom namespaces in queries
        df = self.get_table_and_convert(self.regions_query, index="mrid")
        return RegionsDataFrame(df)

    @property
    def hvdc_query(self) -> str:
        return self.template_to_query(templates.HVDC)

    @time_it
    def hvdc(self) -> HVDCDataFrame:
        df = self.get_table_and_convert(self.hvdc_query)
        return HVDCDataFrame(df)

    @property
    def hvdc_converter_bidzone_query(self) -> str:
        return self.template_to_query(templates.HVDC_CONVERTER_BIDZONES)

    @time_it
    def hvdc_converter_bidzones(self) -> HVDCBidzonesDataFrame:
        """Fetch mrid of converters placed on HVDC exchange corridors together with to/from bidzone."""
        df = self.get_table_and_convert(self.hvdc_converter_bidzone_query, index="mrid")
        return HVDCBidzonesDataFrame(df)

    @property
    def ras_equipment_query(self) -> str:
        return self.template_to_query(templates.RAS_EQUIPMENT_QUERY)

    @time_it
    def ras_equipment(self) -> RASEquipmentDataFrame:
        df = self.get_table_and_convert(self.ras_equipment_query)
        return RASEquipmentDataFrame(df)

    def add_mrid_query(self, rdf_type: str | None = None, graph: str | None = None) -> str:
        substitutes = {"rdf_type": rdf_type or "?rdf_type", "g": graph or "?g"}
        return self.template_to_query(templates.ADD_MRID_QUERY, substitutes)

    def add_mrid(
        self,
        rdf_type: str | None = None,
        graph: str | None = None,
        client: GraphDBClient | None = None,
    ) -> None:
        """Add cim:IdentifiedObject.mRID to all records.

        It is copied from rdf:about (or rdf:ID) if replace is not specified. The query is executed with the passed
        client. If not given, the default client is used.

        Args:
            graph: Name of graph where mrids should be added. Note, mrid is added to all objects
                in the graph.
            rdf_type: RDF type where ID should be added
            client: Client used to execute the query
        """
        client = client or self.client
        client.update_query(self.add_mrid_query(rdf_type, graph))

    def switches_query(self) -> str:
        return self.template_to_query(templates.SWITCHES_QUERY)

    @time_it
    def switches(self) -> SwitchesDataFrame:
        df = self.get_table_and_convert(self.switches_query(), index="mrid")
        return SwitchesDataFrame(df)

    def connectivity_nodes_query(self, region: str | None = None) -> str:
        substitutes = {"region": region or ".*"}
        return self.template_to_query(templates.CONNECTIVITY_NODES_QUERY, substitutes)

    @time_it
    def connectivity_nodes(self, region: str | None = None) -> ConnectivityNodeDataFrame:
        df = self.get_table_and_convert(self.connectivity_nodes_query(region), index="mrid")
        return ConnectivityNodeDataFrame(df)

    def sv_power_deviation_query(self) -> str:
        return self.template_to_query(templates.SV_POWER_DEVIATION_QUERY)

    @time_it
    def sv_power_deviation(self) -> SvPowerDeviationDataFrame:
        return SvPowerDeviationDataFrame(self.get_table_and_convert(self.sv_power_deviation_query()))

    @time_it
    def base_voltage(self) -> BaseVoltageDataFrame:
        query = self.template_to_query(templates.BASE_VOLTAGE)
        return BaseVoltageDataFrame(self.get_table_and_convert(query))

    @time_it
    def associated_switches(self) -> AssociatedSwitchesDataFrame:
        query = self.template_to_query(templates.ASSOCIATED_SWITCHES)
        return AssociatedSwitchesDataFrame(self.get_table_and_convert(query))

    @time_it
    def gen_unit_and_sync_machine_mrid(self) -> GenUnitAndSyncMachineMridDataFrame:
        query = self.template_to_query(templates.GEN_UNIT_MRID_AND_SYNC_MACHINE)
        return GenUnitAndSyncMachineMridDataFrame(self.get_table_and_convert(query))

    @time_it
    def station_group_for_power_unit(self) -> StationGroupForPowerUnitDataFrame:
        query = self.template_to_query(templates.STATION_GROUP_FOR_POWER_UNIT_QUERY)
        return StationGroupForPowerUnitDataFrame(self.get_table_and_convert(query))

    @time_it
    def busbar_section(self) -> Generator[str]:
        """Return set of mrids for all busbar sections in model."""
        query = self.template_to_query(templates.BUSBAR_SECTION)
        return (val["mrid"] for val in self.get_client(query_name(query)).exec_query(query).results.values_as_dict())


class SingleClientModel(Model):
    def __init__(
        self,
        client: GraphDBClient,
        config: ModelConfig | None = None,
        mapper: TypeMapper | None = None,
    ) -> None:
        clients = defaultdict[str, GraphDBClient](lambda: client)
        super().__init__(clients, config, mapper)


def get_cim_model(
    service_cfg: ServiceConfig | None = None,
    model_cfg: ModelConfig | None = None,
    custom_headers: dict[str, str] | None = None,
) -> SingleClientModel:
    """Get cim model.

    Function kept for backward compatibility. Use `get_single_client_model` instead.
    """
    return get_single_client_model(service_cfg, model_cfg, custom_headers)


def get_single_client_model(
    service_cfg: ServiceConfig | None = None,
    model_cfg: ModelConfig | None = None,
    custom_headers: dict[str, str] | None = None,
) -> SingleClientModel:
    """Get a CIM Model.

    Args:
        service_cfg: Configurations for the triple store service
        model_cfg: Configurations for the CIM mode
        async_sparql_wrapper: If True http calls are made via asynchronous requests.
            If False, the native SparqlWrapper sends requests via urllib
        custom_headers: Custom headers to be added to the requests
    """
    return SingleClientModel(GraphDBClient(service_cfg, custom_headers), model_cfg)


def get_federated_cim_model(
    eq_client: GraphDBClient,
    tpsvssh_client: GraphDBClient,
    model_cfg: ModelConfig,
    mapper: TypeMapper | None = None,
) -> Model:
    """Get federated CIM model.

    Return a CIM model where the equipment profile is located in one repo and the topology,
    state variables and steady state hypothesis profile is located in another.

    Args:
        eq_client: Client that executes queries from the EQ repository
        tpsvssh_client: Client that executes queries from the TP/SV/SSH repository
        model_cfg: Mode configurations that provides extra information
        mapper: Type mapper
    """
    clients = defaultdict[str, GraphDBClient](lambda: eq_client)  # By default queries are executed from the EQ repo

    # Setup client based on # Name in the pre-defined queries
    exec_from_tpssvssh = (
        "AC Lines",
        "Base voltage",
        "Branch node withdraw",
        "Bus",
        "Converters",
        "DC Active Power Flow",
        "Disconnected",
        "Exchange",
        "Full model",
        "Loads",
        "Phase tap changer",
        "Power flow",
        "SV branch",
        "Series compensators",
        "SvInjection",
        "Switches",
        "Synchronous machines",
        "Sv power deviation",
        "Transformer branches loss",
        "Transformer branches",
        "Transformer center nodes",
        "Winding transformer angle",
        "Windings",
    )
    for query in exec_from_tpssvssh:
        clients[query] = tpsvssh_client
    return Model(clients, model_cfg, mapper)



import numpy as np

from cimsparql.data_models import PhaseTapChangerDataFrame, PhaseTapChangerSchema


def test_empty_phase_tap_changer():
    schema = PhaseTapChangerSchema.to_schema()
    columns = list(schema.columns.keys())
    phase_tap_changer = PhaseTapChangerDataFrame(columns=columns)

    # Check that the target value is correctly casted from object to np.float64
    assert phase_tap_changer.dtypes.to_dict()["target_value"] == np.float64



import os

import pytest

import tests.t_utils.custom_models as t_custom
from cimsparql.graphdb import GraphDBClient


@pytest.fixture
def gdbc() -> GraphDBClient:
    if not (os.getenv("GRAPHDB_SERVER") and os.getenv("GRAPHDB_TOKEN")):
        pytest.skip("Need graphdb server to run")
    return GraphDBClient(t_custom.combined_graphdb_service())


def test_get_prefixes(gdbc: GraphDBClient):
    for prefix in ["rdf", "cim", "SN"]:
        assert prefix in gdbc.prefixes


def test_prefix_ns(gdbc: GraphDBClient):
    ns = gdbc.prefixes
    assert ns["rdf"] == "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
    assert ns["cim"] == "http://iec.ch/TC57/2013/CIM-schema-cim16#"



from __future__ import annotations

import functools
import os
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from cimsparql.graphdb import GraphDBClient, ServiceConfig, config_bytes_from_template, confpath, new_repo

if TYPE_CHECKING:
    from cimsparql.model import Model

this_dir = Path(__file__).parent


@dataclass
class ModelTest:
    model: Model | None = None
    must_run_in_ci: bool = True
    cleanup: bool = True
    name: str = ""


@functools.lru_cache
def init_repo_rdf4j(url: str, name: str) -> GraphDBClient:
    template = confpath() / "native_store_config_template.ttl"
    config = config_bytes_from_template(template, {"repo": name})
    service_config = ServiceConfig(repo=name, server=url, protocol="http", timeout=30)
    return new_repo(service_config, config, allow_exist=False)


@functools.lru_cache
def initialized_rdf4j_repo() -> GraphDBClient:
    data_path = this_dir.parent / "data"
    client = init_repo_rdf4j(rdf4j_url(), "picasso")

    data_file = data_path / "artist.ttl"
    client.upload_rdf(data_file, "turtle")
    return client


def rdf4j_url() -> str:
    if os.getenv("CI"):
        # Running on GitHub
        return "localhost:8080/rdf4j-server"

    return os.getenv("RDF4J_URL", "")


def blazegraph_url() -> str:
    if os.getenv("CI"):
        # Running on GitHub
        return "localhost:9999/blazegraph/namespace"
    return os.getenv("BLAZEGRAPH_URL", "")


def check_model(test_model: ModelTest) -> None:
    if not test_model.model:
        if test_model.must_run_in_ci and os.getenv("CI"):
            pytest.fail("Micro model test can not be skipped in CI pipelines")
        else:
            pytest.skip("Can not access micro model")



from collections import defaultdict
from http import HTTPStatus
from pathlib import Path
from typing import Any

import pytest
import werkzeug
from pytest_httpserver import HTTPServer
from SPARQLWrapper import SPARQLWrapper

from cimsparql.adaptions import is_uuid
from cimsparql.graphdb import GraphDBClient, RestApi, ServiceConfig
from cimsparql.model import Model, ModelConfig, get_federated_cim_model
from cimsparql.sparql_result_json import SparqlData, SparqlResultHead, SparqlResultJson, SparqlResultJsonFactory
from cimsparql.templates import sparql_folder
from cimsparql.type_mapper import TypeMapper
from cimsparql.utils import query_name


def test_map_data_types():
    config = ServiceConfig(rest_api=RestApi.DIRECT_SPARQL_ENDPOINT)
    client = GraphDBClient(config)
    cim_model = Model({"default": client}, mapper=LocalTypeMapper(client))
    assert cim_model.map_data_types


def test_not_map_data_types():
    config = ServiceConfig(rest_api=RestApi.DIRECT_SPARQL_ENDPOINT)
    client = GraphDBClient(config)
    cim_model = Model({"default": client}, mapper=LocalTypeMapper(client))
    cim_model.client.prefixes.pop("cim")
    assert not cim_model.map_data_types


@pytest.mark.parametrize("sparql_query", sparql_folder.glob("*.sparql"))
def test_name_in_header(sparql_query: Path):
    with sparql_query.open() as infile:
        line = infile.readline()
    assert line.startswith("# Name: ")


def test_unique_headers():
    headers = list[str]()
    for f in sparql_folder.glob("*.sparql"):
        with f.open() as infile:
            headers.append(infile.readline())

    assert len(headers) == len(set(headers))


def test_query_name():
    query = """# Name: name of the query
    select * {?s ?p ?o}
    """
    assert query_name(query) == "name of the query"


def test_query_without_name():
    query = "select * {?s ?p ?o}"
    assert query_name(query) == ""


def test_multi_client_model_defined_clients_exist():
    config = ServiceConfig(rest_api=RestApi.DIRECT_SPARQL_ENDPOINT)
    eq = GraphDBClient(config)
    tpsv = GraphDBClient(config)
    m_cfg = ModelConfig()

    model = get_federated_cim_model(eq, tpsv, m_cfg, LocalTypeMapper(eq))

    query_names = list[str]()
    for f in sparql_folder.glob("*.sparql"):
        with f.open() as infile:
            query_names.append(query_name(infile.read()))

    queries_with_client = set(model.clients.keys())

    # Verify that all queries with a special client is one of the pre-defined queries
    assert queries_with_client.issubset(query_names)


class LocalTypeMapper(TypeMapper):
    def get_map(self) -> dict[str, Any]:
        return {}


class CorrelationIdPicker:
    def __init__(self) -> None:
        self.correlation_id = None

    def extract_correlation_id(self, request: werkzeug.Request) -> werkzeug.Response:
        self.correlation_id = request.headers.get(GraphDBClient.x_correlation_id)
        result = SparqlResultJsonFactory.build().model_dump_json()
        return werkzeug.Response(result, status=HTTPStatus.OK, mimetype="application/json")


def test_correlation_id(httpserver: HTTPServer):
    correlation_picker = CorrelationIdPicker()
    httpserver.expect_request("/sparql").respond_with_handler(correlation_picker.extract_correlation_id)
    config = ServiceConfig(server=httpserver.url_for("/sparql"), rest_api=RestApi.DIRECT_SPARQL_ENDPOINT)

    model = Model(
        {"name1": GraphDBClient(config), "name2": GraphDBClient(config)},
        mapper=LocalTypeMapper(GraphDBClient(config)),
    )
    query1 = "# Name: name1\nselect * {?s ?p ?o}"
    query2 = "# Name: name2\nselect * {?s ?p ?o}"

    model.get_table_and_convert(query1)  # Runs from first client
    assert correlation_picker.correlation_id is None
    model.get_table_and_convert(query2)  # Runs from second client
    assert correlation_picker.correlation_id is None

    # Run from within context
    with model:
        model.get_table_and_convert(query1)  # Runs from first client
        assert correlation_picker.correlation_id
        assert is_uuid(correlation_picker.correlation_id)

        c_id = correlation_picker.correlation_id
        model.get_table_and_convert(query2)  # Runs from second client (but should have same correlation id)
        assert correlation_picker.correlation_id == c_id

    # Run again and confirm that correlation id is removed
    model.get_table_and_convert(query1)  # Runs from first client
    assert correlation_picker.correlation_id is None
    model.get_table_and_convert(query2)  # Runs from second client
    assert correlation_picker.correlation_id is None


class EmptyThreeWindingTransformerSPARQLWrapper(SPARQLWrapper):
    def __init__(self) -> None:
        super().__init__("http://fixed-result-endpoint")
        self.result = SparqlResultJsonFactory.build()

    def three_winding_result(self) -> SparqlResultJson:
        variables = [
            "node_1",
            "node_2",
            "status",
            "name",
            "mrid",
            "un",
            "r",
            "x",
            "b",
            "g",
            "rate",
            "bidzone_1",
            "bidzone_2",
            "angle",
            "ratio",
            "connectivity_node_1",
            "connectivity_node_2",
        ]
        return SparqlResultJson(head=SparqlResultHead(vars=variables), results=SparqlData(bindings=[]))

    def angle(self) -> SparqlResultJson:
        return SparqlResultJson(head=SparqlResultHead(vars=["mrid", "angle"]), results=SparqlData(bindings=[]))

    def three_winding_loss(self) -> SparqlResultJson:
        variables = ["mrid", "ploss_2"]
        return SparqlResultJson(head=SparqlResultHead(vars=variables), results=SparqlData(bindings=[]))

    def queryAndConvert(self) -> dict[str, Any]:  # noqa: N802
        name = query_name(self.queryString)
        if name == "Transformer branches":
            result = self.three_winding_result()
        elif name == "Transformer branches loss":
            result = self.three_winding_loss()
        elif name == "Winding transformer angle":
            result = self.angle()
        else:
            raise ValueError(f"No handler for query: {self.queryString}")
        return result.model_dump(mode="json")


def test_three_winding_empty_result():
    config = ServiceConfig(server="http://some-server", rest_api=RestApi.DIRECT_SPARQL_ENDPOINT)
    client = GraphDBClient(service_cfg=config, sparql_wrapper=EmptyThreeWindingTransformerSPARQLWrapper())
    model = Model(defaultdict(lambda: client), mapper=LocalTypeMapper(client))
    assert model.transformer_branches().empty



"""Adaption functionality used to modify test cases."""

from __future__ import annotations

import hashlib
import re
import uuid
from contextlib import suppress
from typing import TYPE_CHECKING

from pyoxigraph import BlankNode, DefaultGraph, Literal, NamedNode, Quad, QuerySolutions, RdfFormat, Store

from cimsparql.graphdb import default_namespaces

if TYPE_CHECKING:
    from collections.abc import Container, Iterator
    from pathlib import Path


class StandardNamespaces:
    rdf_type = NamedNode("http://www.w3.org/1999/02/22-rdf-syntax-ns#type")
    xsd_integer = NamedNode("http://www.w3.org/2001/XMLSchema#integer")
    xsd_boolean = NamedNode("http://www.w3.org/2001/XMLSchema#boolean")
    xsd_float = NamedNode("http://www.w3.org/2001/XMLSchema#float")


class XmlModelAdaptor:
    eq_predicate = "http://entsoe.eu/CIM/EquipmentCore/3/1"

    def __init__(self, filenames: Iterator[Path]) -> None:
        self.store = Store()
        self.ns = default_namespaces()

        for filename in filenames:
            profile = filename.stem.rpartition("/")[-1]
            uri = NamedNode(f"http://cimsparql/xml-adpator/{profile}")
            profile_store = Store()
            profile_store.load(None, RdfFormat.RDF_XML, path=filename, base_iri="http://cim.example.org#")
            for quad in profile_store.quads_for_pattern(None, None, None, None):
                self.store.add(Quad(quad.subject, quad.predicate, quad.object, uri))

    def select_query(self, query: str, *, prefixes: dict[str, str] | None = None) -> QuerySolutions:
        result = self.store.query(query, prefixes=prefixes)
        assert isinstance(result, QuerySolutions)
        return result

    def namespaces(self) -> dict[str, str]:
        return self.ns

    def graphs(self) -> set[str]:
        return {graph for _, _, _, graph in self.store.quads_for_pattern(None, None, None, None) if graph}

    def all_quads(self) -> Iterator[Quad]:
        return self.store.quads_for_pattern(None, None, None, None)

    @classmethod
    def from_folder(cls, folder: Path) -> XmlModelAdaptor:
        return XmlModelAdaptor(folder.glob("*.xml"))

    def add_mrid(self) -> None:
        """Add cim:IdentifiedObject.mRID if not present."""
        ns = self.namespaces()
        identified_obj_mrid = NamedNode(f"{ns['cim']}IdentifiedObject.mRID")
        has_mrid = set(self.store.quads_for_pattern(None, identified_obj_mrid, None, None))
        missing_mrid = set(self.all_quads()) - has_mrid

        seen_subjects = set[NamedNode | BlankNode]()

        # Sort the quads such that quads in the EQ graph is considered before the other profiles
        for quad in sorted(missing_mrid, key=lambda quad: "EQ" not in str(quad.graph_name)):
            if quad.subject in seen_subjects:
                continue
            assert isinstance(quad.subject, (NamedNode, BlankNode))
            seen_subjects.add(quad.subject)
            mrid_str = str(quad.subject.value).rpartition("#_")[-1]
            mrid = mrid_str if is_uuid(mrid_str) else generate_uuid(mrid_str)
            self.store.add(Quad(quad.subject, identified_obj_mrid, Literal(mrid), quad.graph_name))

    def add_market_code_to_non_conform_load(self) -> None:
        updated = set[NamedNode | BlankNode]()
        for load, _, _, ctx in self.store.quads_for_pattern(
            None, StandardNamespaces.rdf_type, NamedNode(self.ns["cim"] + "EnergyConsumer"), None
        ):
            if load in updated:
                continue

            if "EQ" not in str(ctx):
                continue
            updated.add(load)

            load_group = BlankNode()
            schedule_resource = BlankNode()
            self.store.add(Quad(load, NamedNode(self.ns["cim"] + "NonConformLoad.LoadGroup"), load_group, ctx))
            self.store.add(
                Quad(load_group, NamedNode(self.ns["cim"] + "IdentifiedObject.name"), Literal("created_group"), ctx)
            )
            self.store.add(
                Quad(
                    load_group,
                    NamedNode(self.ns["SN"] + "NonConformLoadGroup.ScheduleResource"),
                    schedule_resource,
                    ctx,
                )
            )
            self.store.add(
                Quad(
                    schedule_resource,
                    NamedNode(self.ns["SN"] + "ScheduleResource.marketCode"),
                    Literal("market001"),
                    ctx,
                )
            )

    def adapt(self, eq_uri: str) -> None:
        self.add_zero_sv_power_flow()
        self.add_zero_sv_injection()
        self.add_generating_unit()
        self.add_market_code_to_non_conform_load()
        self.add_protective_action_equipment()
        self.add_mrid()
        self.add_dtypes()
        self.add_internal_eq_link(eq_uri)
        self.add_eic_code()
        self.add_network_analysis_enable()

    def add_zero_sv_power_flow(self) -> None:
        sv_graph = NamedNode("http://cimsparql/xml-adpator/SV")
        sv_power_flow_terminal = NamedNode(self.ns["cim"] + "SvPowerFlow.Terminal")
        has_sv_power_flow = {
            quad.object for quad in self.store.quads_for_pattern(None, sv_power_flow_terminal, None, None)
        }

        zero_sv_flow_added = set[NamedNode | BlankNode]()

        for terminal, _, _, _ in self.store.quads_for_pattern(
            None, StandardNamespaces.rdf_type, NamedNode(self.ns["cim"] + "Terminal"), None
        ):
            # Do not add SvPowerFlow if it already exists
            if terminal in has_sv_power_flow or terminal in zero_sv_flow_added:
                continue
            zero_sv_flow_added.add(terminal)
            power_flow = BlankNode()
            self.store.add(Quad(power_flow, sv_power_flow_terminal, terminal, sv_graph))
            self.store.add(Quad(power_flow, NamedNode(self.ns["cim"] + "SvPowerFlow.p"), Literal(0.0), sv_graph))
            self.store.add(Quad(power_flow, NamedNode(self.ns["cim"] + "SvPowerFlow.q"), Literal(0.0), sv_graph))

    def add_dtypes(self) -> None:
        fields = {
            "endNumber": int,
            "sequenceNumber": int,
            "phaseAngleClock": int,
            "SvPowerFlow.p": float,
            ".open": to_boolean,
            ".connected": to_boolean,
            ".nominalVoltage": float,
        }
        for quad in self.all_quads():
            with suppress(StopIteration):
                f = next(f for f in fields if f in str(quad.predicate))
                self.store.remove(quad)

                caster = fields[f]
                assert isinstance(quad.object, Literal)
                literal = Literal(caster(quad.object.value))
                self.store.add(Quad(quad.subject, quad.predicate, literal, quad.graph_name))

    def tpsvssh_contexts(self) -> Iterator[NamedNode | BlankNode]:
        return (
            node for node in self.store.named_graphs() if any(substr in str(node) for substr in ("SV", "TP", "SSH"))
        )

    def eq_contexts(self) -> Iterator[NamedNode | BlankNode]:
        return (node for node in self.store.named_graphs() if any(substr in str(node) for substr in ("EQ", "GL")))

    def contexts(self) -> Iterator[NamedNode | BlankNode]:
        return self.store.named_graphs()

    def nq_bytes(self, contexts: Container[NamedNode | BlankNode | DefaultGraph] | None = None) -> bytes:
        """Return the contexts as bytes. If contexts is None, the entire graph is exported."""
        store = Store()
        for quad in self.all_quads():
            if contexts is None or quad.graph_name in contexts:
                store.add(quad)

        result = store.dump(format=RdfFormat.N_QUADS)
        assert isinstance(result, bytes)
        return result

    def add_internal_eq_link(self, eq_uri: str) -> None:
        # Insert in one SV graph
        ctx = next(c for c in self.tpsvssh_contexts() if "SV" in str(c))
        self.store.add(Quad(BlankNode(), NamedNode(self.eq_predicate), NamedNode(eq_uri), ctx))

    def add_zero_sv_injection(self) -> None:
        tp_node_type = NamedNode(self.ns["cim"] + "TopologicalNode")
        sv_inj_type = NamedNode(self.ns["cim"] + "SvInjection")

        graph = NamedNode("http://cimsparql/xml-adpator/SV-injection")
        used_nodes = set[NamedNode | BlankNode]()
        for s, _, _, _ in self.store.quads_for_pattern(None, StandardNamespaces.rdf_type, tp_node_type, None):
            if s in used_nodes:
                continue
            used_nodes.add(s)
            sv_injection = BlankNode()
            self.store.add(Quad(sv_injection, StandardNamespaces.rdf_type, sv_inj_type, graph))
            self.store.add(Quad(sv_injection, NamedNode(self.ns["cim"] + "SvInjection.TopologicalNode"), s, graph))
            self.store.add(
                Quad(sv_injection, NamedNode(self.ns["cim"] + "SvInjection.pInjection"), Literal(0.0), graph)
            )

    def add_eic_code(self) -> None:
        for substation, _, _, ctx in self.store.quads_for_pattern(
            None, StandardNamespaces.rdf_type, NamedNode(self.ns["cim"] + "Substation"), None
        ):
            market_delivery_point = BlankNode()
            bidding_area = BlankNode()
            eic_code = Literal("10Y1001A1001A48H")
            self.store.add(
                Quad(
                    substation, NamedNode(f"{self.ns['SN']}Substation.MarketDeliveryPoint"), market_delivery_point, ctx
                )
            )
            self.store.add(
                Quad(
                    market_delivery_point,
                    NamedNode(f"{self.ns['SN']}MarketDeliveryPoint.BiddingArea"),
                    bidding_area,
                    ctx,
                )
            )
            self.store.add(
                Quad(
                    bidding_area,
                    NamedNode(f"{self.ns['entsoeSecretariat']}IdentifiedObject.energyIdentCodeEIC"),
                    eic_code,
                    ctx,
                )
            )

    def add_network_analysis_enable(self) -> None:
        conducting_equipment = NamedNode(f"{self.ns['cim']}Terminal.ConductingEquipment")
        used_equipment = set[NamedNode | BlankNode]()
        for _, _, equipment, ctx in self.store.quads_for_pattern(None, conducting_equipment, None, None):
            if equipment in used_equipment:
                continue
            used_equipment.add(equipment)
            assert ctx
            network_analysis_enable = NamedNode(f"{self.ns['SN']}Equipment.networkAnalysisEnable")
            self.store.add(Quad(equipment, network_analysis_enable, Literal(value=True), ctx))

    def add_generating_unit(self) -> None:
        updated = set[NamedNode | BlankNode]()
        sync_machine_type = NamedNode(self.ns["cim"] + "SynchronousMachine")
        for sync_machine, _, _, ctx in self.store.quads_for_pattern(
            None, StandardNamespaces.rdf_type, sync_machine_type, None
        ):
            if "EQ" not in str(ctx):
                continue
            if sync_machine in updated:
                continue
            generating_unit = BlankNode()
            schedule_resource = BlankNode()
            updated.add(sync_machine)

            gen_unit_pred = NamedNode(self.ns["cim"] + "SynchronousMachine.GeneratingUnit")
            thermal_gen_unit_type = NamedNode(self.ns["cim"] + "ThermalGeneratingUnit")
            name_pred = NamedNode(self.ns["cim"] + "IdentifiedObject.name")
            schedule_resource_pred = NamedNode(self.ns["SN"] + "GeneratingUnit.ScheduleResource")
            scheduler_resource_type = NamedNode(self.ns["SN"] + "ScheduleResource")
            market_code_pred = NamedNode(self.ns["SN"] + "ScheduleResource.marketCode")

            self.store.add(Quad(sync_machine, gen_unit_pred, generating_unit, ctx))
            self.store.add(Quad(gen_unit_pred, StandardNamespaces.rdf_type, thermal_gen_unit_type))
            self.store.add(Quad(generating_unit, name_pred, Literal("GeneratuinggUnit"), ctx))
            self.store.add(Quad(generating_unit, schedule_resource_pred, schedule_resource, ctx))
            self.store.add(Quad(schedule_resource, StandardNamespaces.rdf_type, scheduler_resource_type, ctx))
            self.store.add(Quad(schedule_resource, market_code_pred, Literal("market001"), ctx))

    def add_protective_action_equipment(self) -> None:
        rpact = BlankNode()
        sync_machine, _, _, _ = next(
            self.store.quads_for_pattern(
                None, StandardNamespaces.rdf_type, NamedNode(self.ns["cim"] + "SynchronousMachine"), None
            )
        )
        eq_graph = next(self.eq_contexts())

        literal_true = Literal("true", datatype=StandardNamespaces.xsd_boolean)
        literal_false = Literal("false", datatype=StandardNamespaces.xsd_boolean)

        self.store.add(
            Quad(rpact, NamedNode(self.ns["ALG"] + "ProtectiveActionEquipment.Equipment"), sync_machine, eq_graph)
        )
        self.store.add(
            Quad(rpact, StandardNamespaces.rdf_type, NamedNode(self.ns["ALG"] + "ProtectiveActionEquipment"), eq_graph)
        )
        self.store.add(Quad(rpact, NamedNode(self.ns["cim"] + "IdentifiedObject.name"), Literal("ras"), eq_graph))
        self.store.add(Quad(rpact, NamedNode(self.ns["ALG"] + "ProtectiveAction.flowShift"), literal_true, eq_graph))
        self.store.add(
            Quad(rpact, NamedNode(self.ns["ALG"] + "ProtectiveAction.flowShiftFlip"), literal_true, eq_graph)
        )
        self.store.add(
            Quad(rpact, NamedNode(self.ns["ALG"] + "ProtectiveAction.loadContribution"), literal_false, eq_graph)
        )
        self.store.add(
            Quad(rpact, NamedNode(self.ns["ALG"] + "ProtectiveAction.unitContribution"), literal_false, eq_graph)
        )


def is_uuid(x: str) -> bool:
    return re.match("^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$", x) is not None


def generate_uuid(x: str) -> str:
    h = hashlib.md5(x.encode(), usedforsecurity=False)
    return str(uuid.UUID(hex=h.hexdigest()))


def to_boolean(x: str) -> bool:
    return x.lower() in {"true", "1"}



"""Graphdb CIM sparql client."""

from __future__ import annotations

import json
import os
from copy import deepcopy
from dataclasses import dataclass, field
from enum import StrEnum
from http import HTTPStatus
from pathlib import Path
from typing import TYPE_CHECKING, TypedDict, TypeVar

import httpx
import pandas as pd
import tenacity
from SPARQLWrapper import JSON, POST, SPARQLWrapper

from cimsparql.retry_cb import RetryCallback, RetryCallbackFactory
from cimsparql.sparql_result_json import SparqlData, SparqlResultHead, SparqlResultJson, SparqlResultValue
from cimsparql.url import service, service_blazegraph

if TYPE_CHECKING:
    from collections.abc import Callable

    from tenacity.stop import stop_base

    from cimsparql.sparql_result_json import SparqlResultValue


class SparqlResult(TypedDict):
    cols: list[str]
    data: list[dict[str, SparqlResultValue]]


def data_row(cols: list[str], rows: list[dict[str, SparqlResultValue]]) -> dict[str, SparqlResultValue]:
    """Get a sample row for extraction of data types.

    Args:
       cols: queried columns (optional might return None)
       rows: 'results'→'bindings' from SPARQLWrapper

    Returns:
       samples result of all columns in query
    """
    full_row: dict[str, SparqlResultValue] = {}
    for row in rows:
        if set(cols).difference(full_row.keys()):
            full_row.update(row)
        else:
            break
    return full_row


class RestApi(StrEnum):
    RDF4J = "RDF4J"
    BLAZEGRAPH = "BLAZEGRAPH"
    DIRECT_SPARQL_ENDPOINT = "DIRECT_SPARQL_ENDPOINT"


def parse_namespaces_rdf4j(response: httpx.Response) -> dict[str, str]:
    return dict(line.split(",") for line in response.text.split()[1:])


@dataclass(frozen=True)
class ServiceConfig:
    repo: str = field(default=os.getenv("GRAPHDB_REPO", "LATEST"))
    protocol: str = "https"
    server: str = field(default=os.getenv("GRAPHDB_SERVER", "127.0.0.1:7200"))
    path: str = ""
    user: str | None = field(default=os.getenv("GRAPHDB_USER"))
    passwd: str | None = field(default=os.getenv("GRAPHDB_USER_PASSWD"))
    token: str | None = field(default=os.getenv("GRAPHDB_TOKEN"))
    rest_api: RestApi = field(default=RestApi(os.getenv("SPARQL_REST_API", "RDF4J")))
    ca_bundle: str | None = field(default=None)
    retry_callback_factory: RetryCallbackFactory = field(default=RetryCallback)
    retry_stop_criteria: stop_base = field(default=tenacity.stop_after_attempt(1))

    # Parameters for rest api
    # https://rdf4j.org/documentation/reference/rest-api/
    distinct: bool = False
    infer: bool = False
    limit: int | None = None
    offset: int | None = None
    timeout: int | None = None
    max_delay_seconds: int = 60
    validate: bool = False

    def __post_init__(self) -> None:
        if self.rest_api not in RestApi:
            raise ValueError(f"rest_api must be one of {RestApi}")

    @property
    def url(self) -> str:
        if self.rest_api == RestApi.BLAZEGRAPH:
            return service_blazegraph(self.server, self.repo, self.protocol)
        if self.rest_api == RestApi.DIRECT_SPARQL_ENDPOINT:
            return self.server
        return service(self.repo, self.server, self.protocol, self.path)

    @property
    def parameters(self) -> dict[str, bool | int | None]:
        return {
            "distinct": self.distinct,
            "infer": self.infer,
            "limit": self.limit,
            "offset": self.offset,
        }

    @property
    def auth(self) -> httpx.BasicAuth | None:
        return httpx.BasicAuth(self.user, self.passwd) if self.user and self.passwd and not self.token else None


# Available formats from RDF4J API
# https://rdf4j.org/documentation/reference/rest-api/
MIME_TYPE_RDF_FORMATS = {
    "rdf/xml": "application/rdf+xml",
    "n-triples": "text/plain",
    "turtle": "text/turtle",
    "n3": "text/rdf+n3",
    "n-quads": "text/x-nquads",
    "json-ld": "application/ld+json",
    "rdf/json": "application/rdf+json",
    "trix": "application/trix",
    "trig": "application/x-trig",
    "rdf4J binary rdf": "application/x-binary-rdf",
}

UPLOAD_END_POINT = {RestApi.RDF4J: "/statements", RestApi.BLAZEGRAPH: ""}

if TYPE_CHECKING:
    from typing import Concatenate, ParamSpec

    P = ParamSpec("P")
    T = TypeVar("T")


def require_rdf4j(
    f: Callable[Concatenate[GraphDBClient, P], T],
) -> Callable[Concatenate[GraphDBClient, P], T]:
    def wrapper(cli: GraphDBClient, *args: P.args, **kwargs: P.kwargs) -> T:
        if cli.service_cfg.rest_api != RestApi.RDF4J:
            raise NotImplementedError("Function only implemented for RDF4J")
        return f(cli, *args, **kwargs)

    return wrapper


class GraphDBClient:
    """GraphDB client for sending sparql queries to GraphDB server.

    Args:
        service_cfg: Service configuration (see ServiceConfig)
        custom_headers: Added to SPARQLWrapper using addCustomHttpHeader
        sparql_wrapper: SPARQLWrapper instance used to post queries. If not given,
        a SPARQLWrapper create posting queries to the URL given in service_cfg

    Example:
    >>> from cimsparql.graphdb import GraphDBClient
    >>> gdbc = GraphDBClient()
    >>> query = 'select * where { ?subject ?predicate ?object } limit 10'
    >>> df, row = gdbc.get_table(query)xs

    Where row is the output of graphdb.data_row
    """

    x_correlation_id = "x-correlation-id"

    def __init__(
        self,
        service_cfg: ServiceConfig | None = None,
        custom_headers: dict[str, str] | None = None,
        sparql_wrapper: SPARQLWrapper | None = None,
    ) -> None:
        self.service_cfg = service_cfg or ServiceConfig()
        self.sparql = sparql_wrapper or SPARQLWrapper(self.service_cfg.url)
        self.sparql.setReturnFormat(JSON)
        self.sparql.setMethod(POST)
        if self.service_cfg.token:
            self.sparql.addCustomHttpHeader("Authorization", self.service_cfg.token)
        else:
            self.sparql.setCredentials(self.service_cfg.user, self.service_cfg.passwd)
        if self.service_cfg.timeout:
            self.sparql.setTimeout(self.service_cfg.timeout)
        self.update_sparql_parameters()
        if custom_headers:
            for name, value in custom_headers.items():
                self.sparql.addCustomHttpHeader(name, value)
        self._prefixes = None

    def add_correlation_id_to_header(self, correlation_id: str) -> None:
        self.sparql.addCustomHttpHeader(self.x_correlation_id, correlation_id)

    def clear_correlation_id_from_header(self) -> None:
        self.sparql.clearCustomHttpHeader(self.x_correlation_id)

    def create_sparql_wrapper(self) -> SPARQLWrapper:
        return SPARQLWrapper(self.service_cfg.url)

    def update_sparql_parameters(self) -> None:
        for key, value in self.service_cfg.parameters.items():
            if value is not None:
                self.set_parameter(key, str(value))

    def set_parameter(self, key: str, value: str) -> None:
        self.sparql.clearParameter(key)
        self.sparql.addParameter(key, value)

    @property
    def prefixes(self) -> dict[str, str]:
        if self._prefixes is None:
            self._prefixes = self.get_prefixes()
        return self._prefixes

    def update_prefixes(self, pref: dict[str, str]) -> None:
        """Update prefixes from a dict."""
        self.prefixes.update(pref)

    def __str__(self) -> str:
        return f"<GraphDBClient object, service: {self.service_cfg.url}>"

    def exec_query(self, query: str) -> SparqlResultJson:
        # To allow exec query to be run in threads, we use a deepcopy of the underlying
        # sparql wrapper .This is needed since setQuery changes the state of the SPARQLWrapper
        sparql_wrapper = deepcopy(self.sparql)
        sparql_wrapper.setQuery(query)

        retry_cb = self.service_cfg.retry_callback_factory()
        retry_cb.pre_call(query)

        sparql_result = None
        for attempt in tenacity.Retrying(
            stop=self.service_cfg.retry_stop_criteria,
            wait=tenacity.wait_exponential(max=self.service_cfg.max_delay_seconds),
            before=retry_cb.before,
            after=retry_cb.after,
        ):
            with attempt:
                results = sparql_wrapper.queryAndConvert()

                sparql_result = SparqlResultJson.model_validate(results)
                if self.service_cfg.validate:
                    sparql_result.validate_column_consistency()
        return sparql_result or SparqlResultJson(head=SparqlResultHead(), results=SparqlData(bindings=[]))

    def _convert_query_result_to_df(
        self, sparql_result: SparqlResultJson
    ) -> tuple[pd.DataFrame, dict[str, SparqlResultValue]]:
        df = (
            pd.DataFrame(sparql_result.results.values_as_dict(), columns=sparql_result.head.variables)
            if len(sparql_result.results.bindings)
            else pd.DataFrame(columns=sparql_result.head.variables)
        )
        return df, data_row(sparql_result.head.variables, sparql_result.results.bindings)

    def get_table(self, query: str) -> tuple[pd.DataFrame, dict[str, SparqlResultValue]]:
        """Get result from sparql query as a pandas dataframe.

        Args:
           query: to sparql server
           limit: limit number of resulting rows

        """
        return self._convert_query_result_to_df(self.exec_query(query))

    @property
    def empty(self) -> bool:
        """Identify empty GraphDB repo."""
        return self.get_table("select * where {?s ?p ?o} limit 1")[0].empty

    def get_prefixes(self, http_transport: httpx.BaseTransport | None = None) -> dict[str, str]:
        prefixes = default_namespaces()

        if self.service_cfg.rest_api in (RestApi.BLAZEGRAPH, RestApi.DIRECT_SPARQL_ENDPOINT):
            # These APis does not expose prefixes. Custom prefixes must be added
            # via `update_prefixes`. By default we load a pre-defined set of prefixes
            return prefixes

        with httpx.Client(transport=http_transport, timeout=5.0) as client:
            response = client.get(
                self.service_cfg.url + "/namespaces",
                auth=self.service_cfg.auth or httpx.USE_CLIENT_DEFAULT,
                headers=self.sparql.customHttpHeaders,
            )
        if response.status_code == HTTPStatus.OK:
            prefixes.update(parse_namespaces_rdf4j(response))
            return prefixes
        msg = (
            "Could not fetch namespaces and prefixes from graphdb "
            "Verify that user and password are correctly set in the "
            "GRAPHDB_USER and GRAPHDB_USER_PASSWD environment variable"
        )
        raise RuntimeError(f"{msg} Status code: {response.status_code} Reason: {response.reason_phrase}")

    def delete_repo(self) -> None:
        endpoint = delete_repo_endpoint(self.service_cfg)
        response = httpx.delete(endpoint, timeout=5.0)
        response.raise_for_status()

    def upload_rdf(self, content: Path | bytes, rdf_format: str, params: dict[str, str] | None = None) -> None:
        """Upload data in RDF format to a service.

        Args:
            content: Filename to the RDF file to upload
            rdf_format: RDF file type (e.g. rdf/xml, rdf/json etc. Consult the RDF4J rest api
                doc for a complete list of available options)
            params: Additional parameters passed to the post request
        """

        def read_xml_content(file: Path) -> bytes:
            with file.open("rb") as infile:
                return infile.read()

        xml_content = read_xml_content(content) if isinstance(content, Path) else content

        response = httpx.post(
            self.service_cfg.url + UPLOAD_END_POINT[self.service_cfg.rest_api],
            content=xml_content,
            params=params,
            headers={"Content-Type": MIME_TYPE_RDF_FORMATS[rdf_format]},
            timeout=5.0,
        )
        response.raise_for_status()

    def update_query(self, query: str) -> None:
        """Pass a query via a post API call."""
        response = httpx.post(
            self.service_cfg.url + UPLOAD_END_POINT[self.service_cfg.rest_api],
            data={"update": query},
            headers=self.sparql.customHttpHeaders | {"Content-Type": "application/x-www-form-urlencoded"},
            auth=self.service_cfg.auth,
            timeout=5.0,
        )
        response.raise_for_status()

    @require_rdf4j
    def set_namespace(self, prefix: str, value: str) -> None:
        response = httpx.put(
            self.service_cfg.url + f"/namespaces/{prefix}",
            content=value,
            headers={"Content-Type": "text/plain"},
            auth=self.service_cfg.auth,
            timeout=5.0,
        )
        response.raise_for_status()

    @require_rdf4j
    def get_namespace(self, prefix: str) -> str:
        response = httpx.get(self.service_cfg.url + f"/namespaces/{prefix}", auth=self.service_cfg.auth, timeout=5.0)

        response.raise_for_status()
        return response.text


@dataclass
class RepoInfo:
    uri: str
    repo_id: str
    title: str
    readable: bool
    writable: bool


def repos(service_cfg: ServiceConfig | None = None) -> list[RepoInfo]:
    """List available repositories."""
    service_cfg = service_cfg or ServiceConfig()

    auth = httpx.USE_CLIENT_DEFAULT
    if service_cfg.user and service_cfg.passwd:
        auth = httpx.BasicAuth(service_cfg.user, service_cfg.passwd)

    url = f"{service_cfg.protocol}://{service_cfg.server}/repositories"

    with httpx.Client(timeout=5.0) as client:
        response = client.get(url, auth=auth, headers={"Accept": "application/json"})
    response.raise_for_status()

    def _repo_info(repo: dict[str, dict[str, str]]) -> RepoInfo:
        def get(key: str, default: str = "") -> str:
            return repo.get(key, {}).get("value", default)

        uri = get("uri")
        repo_id = get("id")
        title = get("title")
        readable = get("readable", "false") == "true"
        writable = get("writable", "false") == "true"
        return RepoInfo(uri, repo_id, title, readable, writable)

    return [_repo_info(repo) for repo in response.json()["results"]["bindings"]]


def new_repo(service_config: ServiceConfig, config: bytes, allow_exist: bool = True) -> GraphDBClient:
    ignored_errors = {HTTPStatus.CONFLICT} if allow_exist else set[HTTPStatus]()
    url = service_config.url
    response = httpx.put(
        url, content=config, headers={"Content-Type": "text/turtle"}, timeout=service_config.timeout or 5.0
    )
    if response.status_code not in ignored_errors:
        response.raise_for_status()

    return GraphDBClient(service_cfg=service_config)


def new_repo_blazegraph(url: str, repo: str, protocol: str = "https", token: str | None = None) -> GraphDBClient:
    template = confpath() / "blazegraph_repo_config.xml"
    config = config_bytes_from_template(template, {"repo": repo})

    response = httpx.post(
        f"{protocol}://{url}",
        content=config,
        headers={"Content-type": "application/xml"},
        timeout=5.0,
    )
    response.raise_for_status()
    return GraphDBClient(ServiceConfig(repo, protocol, url, rest_api=RestApi.BLAZEGRAPH, token=token))


def config_bytes_from_template(template: Path, params: dict[str, str], encoding: str = "utf-8") -> bytes:
    """Replace value in template file with items in params.

    Args:
        template: Path to template file
        params: Dict with key-value pairs where key must be enclode by double curly braces
            in the template file ({{}}). {{key}} will be replaced by value
        encoding: use this encoding when replacing items
    """
    with template.open("rb") as infile:
        data = infile.read()

    for param, value in params.items():
        enc_templ = f"{{{{{param}}}}}".encode(encoding)
        enc_value = value.encode(encoding)
        data = data.replace(enc_templ, enc_value)
    return data


def confpath() -> Path:
    return Path(__file__).parent.parent / "pkg_data"


def default_namespaces() -> dict[str, str]:
    with (confpath() / "namespaces.json").open() as infile:
        return json.load(infile)


def delete_repo_endpoint(config: ServiceConfig) -> str:
    if config.rest_api == RestApi.BLAZEGRAPH:
        # Remove /sparql at the end
        return config.url.rpartition("/")[0]
    return config.url



"""Type mapper functionality for graphdb data."""

from __future__ import annotations

import dataclasses
from contextlib import contextmanager
from decimal import Decimal
from typing import TYPE_CHECKING, Any

import pandas as pd

from cimsparql.templates import TYPE_MAPPER_QUERY

if TYPE_CHECKING:
    from collections.abc import Callable, Generator

    from cimsparql.graphdb import GraphDBClient

# Type-caster that can be used with pandas. It can be python types, numpy dtypes or string
# value. Examples: float, "int", int, "string"
if TYPE_CHECKING:
    TYPE_CASTER = Callable[[Any], Any] | str
    SPARQL_TYPE = str
    COL_NAME = str
    PREFIX = str
    URI = str

as_type_able = [int, float, "Int64", "Int32", "Int16"]


def to_timedelta(duration: str) -> pd.Timedelta:
    # Pandas only supports days, hours, minutes, seconds
    # not year and month which can be part of
    # https://www.w3.org/TR/xmlschema11-2/#duration

    if "Y" in duration:
        raise ValueError("Cimsparql uses pandas to convert duration. Y not supported")

    return pd.to_timedelta(duration)


def str_preserve_none(x: Any) -> str | None:
    return None if pd.isna(x) else str(x)


XSD_TYPE_MAP: dict[str, TYPE_CASTER] = {
    # Primitive types (https://www.w3.org/TR/xmlschema11-2/#built-in-primitive-datatypes)
    "boolean": lambda x: x.lower() in {"true", "1"},
    "date": pd.to_datetime,
    "dateTime": pd.to_datetime,
    "decimal": Decimal,
    "double": float,
    "duration": to_timedelta,  # Require ISO 8601 format,
    "float": float,
    "integer": int,
    "time": pd.to_datetime,
}


CIM_TYPE_MAP: dict[str, TYPE_CASTER] = {
    "String": str_preserve_none,
    "Integer": int,
    "Boolean": bool,
    "Float": float,
    "Date": pd.to_datetime,
}

sparql_type_map = {"literal": str_preserve_none, "uri": str_preserve_none}


@contextmanager
def enforce_no_limit(client: GraphDBClient) -> Generator[GraphDBClient, None, None]:
    orig_cfg = client.service_cfg
    client.service_cfg = dataclasses.replace(orig_cfg, limit=None)
    client.update_sparql_parameters()

    try:
        yield client
    finally:
        client.service_cfg = orig_cfg
        client.update_sparql_parameters()


def build_type_map(prefixes: dict[PREFIX, URI]) -> dict[SPARQL_TYPE, TYPE_CASTER]:
    short_map = {"xsd": XSD_TYPE_MAP, "cim": CIM_TYPE_MAP}

    type_map: dict[SPARQL_TYPE, TYPE_CASTER] = {}
    for namespace, prim_types in short_map.items():
        if namespace not in prefixes:
            continue
        prefix = prefixes[namespace]
        new_map = {f"{prefix}{dtype}": converter for dtype, converter in prim_types.items()}
        type_map.update(new_map)
    return type_map


class TypeMapper:
    def __init__(
        self,
        client: GraphDBClient,
        custom_additions: dict[str, Any] | None = None,
    ) -> None:
        self.client = client

        self.query = TYPE_MAPPER_QUERY.substitute(self.client.prefixes)
        custom_additions = custom_additions or {}
        self.prim_type_map = build_type_map(self.client.prefixes)
        self.map = sparql_type_map | self.get_map() | self.prim_type_map | custom_additions

    def have_cim_version(self, cim: str) -> bool:
        return any(cim in val for val in self.map)

    def type_map(self, df: pd.DataFrame) -> dict[str, Any]:
        return {
            str(row.sparql_type): self.prim_type_map.get(str(row.range), str_preserve_none) for row in df.itertuples()
        }

    def get_map(self) -> dict[str, Any]:
        """Read all metadata from the sparql backend & creates a sparql-type -> python type map.

        Returns:
            sparql-type -> python type map
        """
        with enforce_no_limit(self.client) as c:
            res = c.get_table(self.query)
            df = res[0]
        if df.empty:
            return {}
        return self.type_map(df)

    def build_type_caster(self, col_map: dict[COL_NAME, SPARQL_TYPE]) -> dict[COL_NAME, TYPE_CASTER]:
        """Construct a direct mapping from column names to a type caster from the."""
        return {col: self.map[dtype] for col, dtype in col_map.items() if dtype in self.map}

    def map_data_types(self, df: pd.DataFrame, col_map: dict[COL_NAME, SPARQL_TYPE]) -> pd.DataFrame:
        """Map the dtypes of a DataFrame to the python-corresponding types of the sparql-types from the source data.

        Args:
            df: DataFrame with columns to be converted
            col_map: type mapper for columns

        Returns:
            mapped DataFrame

        """
        if df.empty:
            return df
        type_caster = self.build_type_caster(col_map)
        df = map_base_types(df, type_caster)
        return map_exceptions(df, type_caster)


def map_base_types(df: pd.DataFrame, type_map: dict[COL_NAME, TYPE_CASTER]) -> pd.DataFrame:
    """Map the datatypes in type_map which can be used with the df.astype function.

    Args:
        df: DataFrame with columns to be converted
        type_map: type mapper for columns

    Returns:
        mapped DataFrame

    """
    as_type_able_columns = {c for c, datatype in type_map.items() if datatype in as_type_able}
    if not df.empty:
        df = df.astype({c: type_map[c] for c in as_type_able_columns})  # type: ignore[reportArgumentType]
    return df


def map_exceptions(df: pd.DataFrame, type_map: dict[COL_NAME, TYPE_CASTER]) -> pd.DataFrame:
    """Map the functions/datatypes in type_map which cant be done with the df.astype function.

    Args:
        df: DataFrame with columns to be converted
        type_map: type mapper for columns

    Returns:
        mapped DataFrame

    """
    ex_columns = {c for c, datatype in type_map.items() if datatype not in as_type_able}
    for column in ex_columns:
        df[column] = df[column].apply(type_map[column])  # type: ignore[reportArgumentType]
    return df
